% GNU-Prolog library for Interaction and Graphical User Interfaces
% Copyright (C) 2001 INRIA Francois Fages
% Version alpha
% License LGPL
% to use with the graphical user interface program CLPgui.java
% http://contraintes.inria.fr/~fages/CLPgui


% Predicates to use
%%%%%%%%%%%%%%%%%%%
% go     to start the CLP server on local host on port 27311
% create_server(Host, Port) creates a socket connection with the GUI on Host
% create_server(Port) creates a socket connection with the GUI on local host
% create_server creates a socket connection with the GUI on local host on port 27311
% create_server(user) command line mode 
% varnames(VarList, NameList) names the variables for the GUI (or the user)
% varnames(VarList) names the variables in VarList 'V1', 'V2', etc
% button(Goal) creates a button in the GUI for executing Goal on demand
% bagof_buttons(Template, Goal) creates a list of buttons in the GUI for each
%                             instance of template satisfying Goal
% show_domain(size)  visualizes the size of FD variable domains
% show_domain(values)  visualizes the values of FD variable domains
% show_domain(interval)  visualizes FD variable domain intervals
% interact executes repeatedly the visualization of domains of the variables
%          and the next command received from the GUI (or the user) in an infinite loop
% disconnect
%
% Command lines
%%%%%%%%%%%%%%%
% 1) Server CLP to Client GUI:
% variables /nl <numberN> /nl <name1> /nl ... <nameN> /nl
% goal /nl <goal> /nl
% undo /nl goal /nl <goal> /nl
% domainSizes /nl <size1> /nl ... <sizeN> /nl
% undo /nl domainSizes /nl
% domainValues /nl <size1> /nl <v11> /nl ... <v1size1> /nl ... <sizeN> /nl <vN1> /nl ... <VNsizeN> /nl
% undo /nl domainValues /nl
% domainIntervals /nl <size1> /nl <min1> /nl <max1> /nl ... <sizeN> /nl <minN> /nl <maxN> /nl
% undo /nl domainValues /nl
% showValues /nl
% showSize /nl
% showInterval /nl
%
% 2) Client GUI to Server CLP
% backtrack /nl
% backtrackInteraction /nl
% <goal> /nl
% showValues /nl
% showSize /nl
% showInterval /nl
% disconnect /nl
%

% These command lines and the GUI will/could be extended to support:
% - visualization of and interaction on traces of execution and constraint propagation
% - visualization of and interaction on other constraint domains (Herbrand)
% - visualization of and interaction on the search tree
% - other CLP servers by porting the program interact.pl to other C(L)P languages

% Bugs:
%
% bagof_buttons doesn't work if it is called with FD variables 
%               (you must call bagof_buttons before fd_domain)
%               this bug of bagof has been corrected in version 1.2.8 of GNU-Prolog


% Examples
%%%%%%%%%%
go:- create_server, interact.

% SEND+MORE=MONEY
sendmore(L):-
	L=[S,E,N,D,M,O,R,Y],
	varnames(L,['S','E','N','D','M','O','R','Y']),
	show_domain(values),
        fd_domain(L,0,9),
        button(fd_domain([S,M],1,9)),
        button(1000*S+100*E+10*N+D + 1000*M+100*O+10*R+E #= 10000*M+1000*O+100*N+10*E+Y),
        button(fd_all_different(L)),
        button(fd_labeling(L)),
	interact.

% Time tabling (Gondran Minoux)
gondran(L):-
	L=[A,B,C,D,E,F,G,H,I,J,K],
	show_domain(values),
	varnames(L,['A','B','C','D','E','F','G','H','I','J','K']),
	%bagof_buttons(fd_labeling(X),member(X,L)),
	fd_domain(L,1,4),
	button(fd_all_different([A,J])),button(fd_all_different([J,I])),button(fd_all_different([I,E])),
	button(fd_all_different([E,C])),button(fd_all_different([C,F])),button(fd_all_different([F,G])),
	button(fd_all_different([D,H])),button(fd_all_different([B,D])),button(fd_all_different([K,E])),
	button(fd_all_different([B,I,H,G])),button(fd_all_different([A,G,E])),
	button(fd_all_different([B,H,K])),button(fd_all_different([A,B,C,H])),
	button(fd_all_different([D,F,J])),
	button(J#>E), button(K#>D), button(K#>F),
	button(fd_atmost(3,L,1)),
	button(fd_atmost(3,L,2)),
	button(fd_atmost(3,L,3)),
	button(fd_atmost(3,L,4)),
	button(fd_labeling(L)),
	interact.


% N queens
queens(N,L):- N>10, !, big_queens(N,L).
queens(N,L):-
	length(L,N),
	varnames(L),
	show_domain(values),
	bagof_buttons(fd_labeling(X),member(X,L)), % bug GNU Prolog < 1.2.8 if bagof is after fd_domain...
	fd_domain(L,1,N),
        button(safe(L)),
        button(fd_labeling(L,[variable_method(ff),value_method(middle)])),
	interact.

big_queens(N,L):- % more appropriate GUI for large number of queens
	length(L,N),
	varnames(L),
	show_domain(size),
	fd_domain(L,1,N),
        button(safe(L)),
        button(fd_labeling(L,[variable_method(ff),value_method(middle)])),
	interact.
 
safe([]).
safe([X| Y]) :-
        noattack(X, Y),
        safe(Y).
 
noattack(X, Xs) :-
        noattack(X, Xs, 1).
 
noattack(_, [], _) :- !.
noattack(X, [Y | Ys], Nb) :-
        X#\=Y,
        X#\=Y+Nb,
        X#\=Y-Nb,
        Nb1 is Nb+1,
        noattack(X, Ys, Nb1).  
	
% Predicates for visualization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

show_domain(size):- !, g_assign(showOption, size), swrite(showSize).
show_domain(values):- !, g_assign(showOption, values), swrite(showValues).
show_domain(interval):- !, g_assign(showOption, interval), swrite(showInterval).
show_domain(O):- write('WARNING unknown option: show_domain '), write(O), nl, g_assign(showOption, size), swrite(showSize).

show_domains:-
	(g_read(interact_varnames, i(_,L)) -> true ; L=[]),
	g_read(showOption,O),
	show_domains(O, L).

show_domains(values, L):- !,
	(swrite(domainValues), showDomains(L); swrite(undo), swrite(domainValues), fail).
show_domains(interval, L):- !,
	(swrite(domainIntervals), showIntervals(L); swrite(undo), swrite(domainIntervals), fail).
show_domains(_, L):- 
	(swrite(domainSizes), showSizes(L); swrite(undo), swrite(domainSizes), fail).

showSizes([]).
showSizes([X|L]):- showSizes(L), fd_size(X,S), swrite(S).

showDomains([]).
showDomains([X|L]):- showDomains(L), fd_size(X,S), swrite(S), fd_dom(X,D), swrite_list(D).
 
showIntervals([]).
showIntervals([X|L]):- showIntervals(L), fd_size(X,S), swrite(S), fd_min(X,M), swrite(M), fd_max(X,N), swrite(N).



% predicates for associating names to variables
varnames(LV):-
	interact_varnaming(LV,LN),
	varnames(LV,LN).

varnames(LV,LN):-
	g_link(interact_varnames,i(LN,LV)).

addvarnames(LV,LN):-
	(g_read(interact_varnames,i(L1,L2)) -> true; L1=[], L2=[]),
	append(LN,L1,R1), append(LV,L2,R2),
	g_link(interact_varnames,i(R1,R2)).

varname(V,N):- var(N), !,
	g_read(interact_varnames,i(LN,LV)), nthq(I,LV,V), nth(I,LN,N).
varname(V,N):- 
	g_read(interact_varnames,i(LN,LV)), nth(I,LN,N), nth(I,LV,V).

nthq(1,[X|_],Y):- X==Y,!.
nthq(I,[_|L],Y):- nthq(J,L,Y), I is J+1.

interact_varnaming(LV,LN):-
	interact_varnaming(LV,LN,0).
interact_varnaming([],[],_).
interact_varnaming([_|LV],[VN|LN],N):-
	number_atom(N,AN),
	atom_concat('V',AN,VN),
	N1 is N+1,
	interact_varnaming(LV,LN,N1).


% button(Goal) creates a button for executing Goal

button(G):-
	goal_to_name(G,N),
	(swrite(goal), swrite(N) ; swrite(undo), swrite(goal), swrite(N), fail).

bagof_buttons(Template,Goal):-
	bagof(Template,Goal,L),
	button_list(L).
button_list([]).
button_list([G|L]):- call(button(G)), button_list(L).

goal_to_name(V,N):- generic_var(V), !,
	(varname(V,N) -> true ;
	    write('WARNING: variable '),write(V),
	    write(' has no name'),nl,N='_').
goal_to_name([],[]):- !.
goal_to_name([T|L],[N|K]):- !,
	goal_to_name(T,N),
	goal_to_name(L,K).
goal_to_name(T,N):-
	T=..[F|L],
	goal_to_name(L,K),
	N=..[F|K].

name_to_goal(N,_):- generic_var(N), !.
name_to_goal('$VARNAME'(N),V):- !,
	((g_read(interact_varnames,i(LN,LV)),nth(I,LN,N)) -> nth(I,LV,V) ; true).
name_to_goal(N,V):- varname(V,N), !.
name_to_goal([],[]):- !.
name_to_goal([N|K],[T|L]):- !,
	name_to_goal(N,T),
	name_to_goal(K,L).
name_to_goal(N,T):-
	N=..[F|K],
	name_to_goal(K,L),
	T=..[F|L].

% interaction engine

interact:- catch(interact1,_,interact).

interact1:- 
        (g_read(interact_varnames,i(L,_))
    ->
        length(L,N), 
        swrite(variables), 
	swrite(N),
	swrite_list(L)
    ;
        true), 
	interact2.
interact1:- interact1. % prevents backtracking before interact

interact2:-
	show_domains, %%%%%%%% VISUALIZATION %%%%%%%%%
	sread(N),     %%%%%%%% INTERACTION   %%%%%%%%%
	(N==backtrack -> g_assign(interactive_backtrack,true), fail;
	    (N==backtrackInteraction -> throw(backtrackInteraction); % undo goals will be lost...
		catch((execute(N), g_assign(interactive_backtrack, false), interact2),
		      backtrackInteraction,
		      (swrite(undo), swrite(goal), swrite(N), fail)))).
interact2:- g_read(interactive_backtrack, true), !, g_assign(interactive_backtrack, false), fail.
interact2:- interact2.

execute(V):- generic_var(V), !, write('ERROR: executing '),write(V),nl.
execute(clear):- !, g_assign(interact_varnames,i([],[])), throw(clear).
execute(N):-
	name_to_goal(N,G),
	%write('calling '), write(G), nl,
	(catch(G,X,write(X)); swrite(undo), swrite(goal), swrite(N), fail).

% Server connexion
create_server:- create_server(_,27311).

create_server(user):- !,
	disconnect,
	g_assign(interact_streams, i(user_input, user_output)).
create_server(Port):-
	create_server(_,Port).

create_server(Host,Port):-
	catch((socket('AF_INET',NbSocket),
	       socket_bind(NbSocket,'AF_INET'(Host,Port)),
               socket_listen(NbSocket,5),
               socket_accept(NbSocket,_,I,O),
	       g_assign(interact_streams, i(I, O))),
	      _,true).
create_server(_,_):- disconnect.

disconnect:-
	g_read(interact_streams,i(I,O)),
	catch((close(I),close(O)),_,true),
	abort.
disconnect.

sread(N):-
	g_read(interact_streams,i(I,_)),
	catch((read_term(I,R,[variable_names(VNL)]),name_query_vars(VNL,_)), X, write(X)),
	%write('READ '), write(R), nl, flush_output,
	(R==showValues -> g_assign(showOption,values), sread(N) ;
	    (R==showSize -> g_assign(showOption, size), sread(N) ;
		(R==showInterval -> g_assign(showOption, interval), sread(N) ;
		    N=R))).
swrite(N):-
	g_read(interact_streams,i(_,O)),
	%write('WRITE '), write(N), nl, flush_output,
	write(O,N),
	snl,
	flush_output(O).
snl:-
	g_read(interact_streams,i(_,O)),
	nl(O),
	flush_output(O).

swrite_list([]).
swrite_list([X|L]):-
	swrite_list(L),
	swrite(X).












