/*-------------------------------------------------------------------------*
 * GNU Prolog                                                              *
 *                                                                         *
 * Part  : CLP(R) Engine                                                   *
 * File  : clpr_pl.pl                                                      *
 * Descr.: CLP(R) solver interface                                         *
 * Author: Remy Haemmerle                                                  *
 *                                                                         *
 * Copyright (C) 2002-2003 INRIA, Remy Haemmerle                           *
 *                                                                         *
 * GNU Prolog is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License as published by the   *
 * Free Software Foundation; either version 2, or any later version.       *
 *                                                                         *
 * GNU Prolog is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of              *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU        *
 * General Public License for more details.                                *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc.  *
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.                     *
 *-------------------------------------------------------------------------*/

'$use_clpr'.

:-built_in.


/* *** Main Predicates *** */

{X}:- set_bip_name({}, 1), clpr_submit(X).

sup(Exp, Inf):-
    set_bip_name(inf,2),
    clpr_minimize(Exp, Inf).

inf(Exp, Sup):-
    set_bip_name(sup,2),
    clpr_minimize(-Exp, Inf),
    Sup is -Inf.

/* *** Miscellaneous Tools Predicates *** */


clpr_epsilon(1.0e-8).

clpr_zero(Z):-
    clpr_epsilon(E),  abs(Z) < E.

clpr_non_zero(Z):-
   clpr_zero(Z) -> fail; true. 

clpr_non_integer(F):-
    clpr_non_zero(F-round(F)).

clpr_integer(F, I):-
    I is round(F), 
    clpr_zero(F-I).


finite(N):-'$call_c_test'('Finite'(N)).


/* *** Normal Forms Traitement *** */


clpr_normalize(X, M, Norm):-generic_var(X), !, Norm = [[X**1.0]*M].
clpr_normalize(X, M, Norm):-ground(X), !, ( 
    C is X*M, clpr_non_zero(C) -> Norm = [[]*C] ; Norm = []).
clpr_normalize(+X, M, Norm):-!, clpr_normalize(X, M, Norm).
clpr_normalize(-X, M, Norm):-!, M1 is -M, clpr_normalize(X, M1, Norm).
clpr_normalize(X+Y, M, Norm):-!,
    clpr_normalize(X, M, NormX),
    clpr_normalize(Y, M, NormY),
    clpr_add_normal_form(NormX, NormY, Norm).
clpr_normalize(X-Y, M, Norm):-!,
    clpr_normalize(X, M, NormX),
    M1 is -M, clpr_normalize(Y, M1, NormY),
    clpr_add_normal_form(NormX, NormY, Norm).
clpr_normalize(X*Y, M, Norm):-!,
    clpr_normalize(X, M, NormX),
    clpr_normalize(Y, 1.0, NormY),
    clpr_mult_normal_form(NormX, NormY, Norm).
clpr_normalize(X/Y, M, Norm):-!,
    clpr_normalize(X, M, NormX),
    clpr_normalize(Y, 1.0, NormY),
    clpr_power_normal_form(NormY, -1.0, NormY2),
    clpr_mult_normal_form(NormX, NormY2, Norm).
clpr_normalize(X**I, M, Norm):-!,(
    ground(I) -> ( 
        I1 is float(I), clpr_non_zero(I) -> 
	    clpr_normalize(X, M, Norm1), 
            clpr_power_normal_form(Norm1, I1, Norm)
        ;
            Norm=[[]*M])
    ;
        throw('non evaluable expression'(_16))).
clpr_normalize(abs(X), M, [[abs(NF)**1.0]*M]):-!, clpr_normalize(X,1.0,NF).
clpr_normalize(sin(X), M, [[sin(NF)**1.0]*M]):-!, clpr_normalize(X,1.0,NF).
clpr_normalize(cos(X), M, [[cos(NF)**1.0]*M]):-!, clpr_normalize(X,1.0,NF).
%clpr_normalize(tan(X), M, [[tan(X)**1.0]*M]):-!, clpr_normalize(X,1.0,NF).
clpr_normalize(min(A,B), M, [[min(NFA,NFB)**1.0]*M]):-!,
    clpr_normalize(A,1.0,NFA),  clpr_normalize(B,1.0,NFB).
clpr_normalize(max(A,B), M, [[max(NFA,NFB)**1.0]*M]):-!,
    clpr_normalize(A,1.0,NFA),  clpr_normalize(B,1.0,NFB).

clpr_normalize(X, _, _):-throw(error('non numeric expression'(X))).

clpr_add_normal_form([], N, N):-!.
clpr_add_normal_form(N, [], N):-!.
clpr_add_normal_form(N1, N2, Norm):-
    N1 = [H1|T1], H1 = K1 * M1, 
    N2 = [H2|T2], H2 = K2 * M2, !, (
    K1 @< K2 -> clpr_add_normal_form(T1, N2, N), Norm=[H1|N];
    K1 @> K2 -> clpr_add_normal_form(T2, N1, N), Norm=[H2|N];
    clpr_add_normal_form(T1, T2, N), 
        (M is M1 + M2, clpr_non_zero(M) -> Norm=[K1*M|N] ; Norm = N)).

clpr_mult_normal_form([], _, []):-!.
clpr_mult_normal_form([H1|T1], N2, Norm):-
    clpr_mult_normal_form(T1, N2, Norm1),
    clpr_mult_normal_form_b(H1, N2, Norm2),
    clpr_add_normal_form(Norm1, Norm2, Norm).

clpr_mult_normal_form_b(_, [], []):-!.
clpr_mult_normal_form_b(H1, [H2|T2], Norm):-
    H1 = K1 * M1, H2 = K2 * M2, 
    clpr_mult_normal_form_b(H1, T2, Norm1),
    clpr_mult_normal_form_c(K1, K2, Norm2),(
    M is M1 * M2, clpr_non_zero(M) ->
      clpr_add_normal_form(Norm1, [Norm2*M], Norm);
      Norm = Norm1).
		      
clpr_mult_normal_form_c([], N, N):-!.
clpr_mult_normal_form_c([H1|T1], N2, Norm):-
    	clpr_mult_normal_form_d(H1, N2, Norm1),
	clpr_mult_normal_form_c(T1, Norm1, Norm).		

clpr_mult_normal_form_d(K, [], [K]):-!.
clpr_mult_normal_form_d(K1, N2, Norm):-
    K1=(V1**I1), N2 = [K2|T2], K2=(V2**I2),(
        V1 @< V2 -> Norm = [K1|N2];
        V1 @> V2 -> clpr_mult_normal_form_d(K1, T2, Norm1), Norm = [K2|Norm1];
        I is I1 + I2, clpr_non_zero(I) -> Norm = [V1**I|T2] ; Norm = T2).

clpr_power_normal_form(_,  I, []):-
    clpr_zero(I), !.
clpr_power_normal_form([], I, []):-!,( 
    I < 0.0 -> fail;
    true ).
clpr_power_normal_form(FN1, I, FN2):-
    clpr_integer(I, I1), I1 > 0 ->
        clpr_power_normal_form_b(FN1, I1, FN2);
        FN2=[[FN1**I]*1.0].

clpr_power_normal_form_b([K1*M1], I, [K2*M2]):-!,
    M2 is M1 ** I,
    clpr_power_normal_form_c(K1, I, K2).

clpr_power_normal_form_b(L, 1, L):-!.
clpr_power_normal_form_b(L, N, L2):-
    N2 is N - 1,
    clpr_power_normal_form_b(L, N2, L1),
    clpr_mult_normal_form(L, L1, L2).


clpr_power_normal_form_c([], _, []):-!.
clpr_power_normal_form_c([K**I1|T1], I, FN):-
    clpr_power_normal_form_c(T1, I, T2), (
        I2 is I1 * I, clpr_non_zero(I2), !, FN=[K**I2|T2];
        FN=T2).


clpr_update_normal_form(NF1, NF3):-
    clpr_update_monoms(NF1,NF2),
    clpr_update_normal_form_b(NF2,NF3).

clpr_update_normal_form_b([H|T], NF):-!,
    clpr_update_normal_form(T,NF1),
    clpr_add_normal_form([H], NF1, NF).
clpr_update_normal_form_b([],[]).

clpr_update_monoms([H1|T1], L2):- ( 
        H1= []*_ -> L2=[H1|T2];  
        H1=K*M1, clpr_update_monom(K, K1), sort0(K1), clpr_update_monom(K1,M1,K2,M2) -> L2=[K2*M2|T2]; 
	L2=T2),
    clpr_update_monoms(T1,T2).
clpr_update_monoms([],[]).  

clpr_update_monom([H1|T1], [H2|T2]):-!,
    clpr_update_monom(T1,T2),
    (H1=(V1**I), clpr_try_flat(V1,V2), H2=(V2**I), !; H1=H2).
 clpr_update_monom([],[]).   

clpr_update_monom([A1|T1], M1, K2, M2):-!,
    A1 = (V1**I1), clpr_update_monom(V1, I1, A1, T1, M1, K2, M2).
clpr_update_monom(X,M,X,M).

clpr_update_monom(C, I1, _, K1, M1, K2, M2):- float(C), !,
    M3 is M1 * C ** I1, clpr_non_zero(M3), clpr_update_monom(K1, M3, K2, M2).

clpr_update_monom(V1, I1, A1, [A2|T2], M1, K2, M2):-!, A2 = (V2 ** I2), (
    V2==V1 -> (
        I is I1 +I2, clpr_non_zero(I) -> 
             clpr_update_monom(V1, I, V1**I, T2, M1, K2, M2);
	     clpr_update_monom(T2, M1, K2, M2));
         clpr_update_monom(V2, I2, A2, T2, M1, K, M2), K2 = [A1|K]).

clpr_update_monom(_ , _ , A1, [], M, [A1], M).

clpr_try_flat(C,C):-(float(C); generic_var(C)), !.
clpr_try_flat(Exp, Out):-Exp=[_|_], !,
  clpr_update_normal_form(Exp, NF1),
  (NF1=[[]*Out], !; Out = NF1).
clpr_try_flat(abs(NF), Out):-!,clpr_update_normal_form(NF, NF1),(
  NF1=[[]*C] -> Out is abs(C);
  NF1=[] -> Out=0.0;
  Out=abs(NF1)).
clpr_try_flat(sin(NF), Out):-!,clpr_update_normal_form(NF, NF1),(
  NF1=[[]*C] -> Out is sin(C);
  NF1=[] -> Out=0.0;
   Out=sin(NF1)).
clpr_try_flat(cos(NF), Out):-!,clpr_update_normal_form(NF, NF1),(
  NF1=[[]*C] -> Out is cos(C);
  NF1=[] -> Out=1.0;
  Out=cos(NF1)).
clpr_try_flat(min(NFA, NFB), Out):-!,
    clpr_update_normal_form(NFA, NFA1), clpr_update_normal_form(NFB, NFB1),(
    (NFA1=[[]*A]; NFA1=[], A=0.0), (NFB1=[[]*B]; NFB1=[], B=0.0) -> Out is min(A,B); Out=min(NFA1, NFB1)).
clpr_try_flat(max(NFA, NFB), Out):-!,
    clpr_update_normal_form(NFA, NFA1), clpr_update_normal_form(NFB, NFB1),(
    (NFA1=[[]*A]; NFA1=[], A=0.0), (NFB1=[[]*B]; NFB1=[], B=0.0) -> Out is max(A,B); Out=max(NFA1, NFB1)).


    
clpr_linear_normal_form([H|T]):-
    H=[X**1.0]*_,!, generic_var(X), clpr_linear_normal_form(T);
    H=[]*_,!, clpr_linear_normal_form(T).
clpr_linear_normal_form([]).

/* *** Submit predicates *** */

clpr_submit(X):-var(X),!,throw(instantition_error).
clpr_submit((X;Y)):-!,(clpr_submit(X);clpr_submit(Y)).
clpr_submit((X,Y)):-!,clpr_submit(X),clpr_submit(Y).
clpr_submit(X=Y):-!,  clpr_normalize(X-Y, 1.0, FN), clpr_submit_eq(FN),!.
clpr_submit(X=<Y):-!, clpr_normalize(X-Y, 1.0, FN), clpr_submit_le(FN),!.
clpr_submit(X>=Y):-!, clpr_normalize(X-Y, -1.0, FN), clpr_submit_le(FN),!.
clpr_submit(X):-'$pl_err_type'('clp(R) constraint',X).



%%%  EQUATIONS



clpr_submit_eq([]):-!.
clpr_submit_eq([A]):-!, clpr_submit_eq_a(A).
clpr_submit_eq([A,B]):-!, clpr_submit_eq_b(A, B).
clpr_submit_eq(Exp):-clpr_submit_eq_c(Exp).

clpr_submit_eq_a([]*_):-!, fail.
clpr_submit_eq_a([K**I]*_):-
    generic_var(K), !, I > 0, K=0.0;
    inverse(K, 0, X, R), !, clpr_submit_eq([[]*R|X]).
clpr_submit_eq_a(Exp):-clpr_submit_eq_c(Exp).

clpr_submit_eq_b([]*C, [X**I]*M):-!, (   
    clpr_integer(I, I1), 0 is (I1 mod 2) -> (
        -C/M < 0.0 -> 
            fail
        ;
            term_variables(X, V), 
            clpr_freeze(V, clpr_submit_eq(_,[[]*C, [X**I]*M])))
    ;
        C2 is -((-C/M) ** (1/I)), finite(C2), (
           generic_var(X) -> X is - C2;
           list(X) -> 
               clpr_add_normal_form([[]*C2], X, N), clpr_submit_eq(N);
           inverse(X, C2, Y, R) ->   
               clpr_submit_eq([[]*R|Y]))).

clpr_submit_eq_b([X1**1.0]*M2, [X2**1.0]*M1):-
    clpr_non_zero(M1+M2) -> fail;
    generic_var(X1), generic_var(X2), X1=X2.
clpr_submit_eq_b(A,B):-clpr_submit_eq_c([A,B]).

clpr_submit_eq_c(Exp):- 
    clpr_linear_normal_form(Exp) -> 
        clpr_submit(0,Exp), 
        '$clpr_try_to_detect_instanciation'
    ;
        term_variables(Exp, V), 
        clpr_freeze(V, clpr_submit_eq(_,Exp)).


clpr_submit_eq(Lock, _):-nonvar(Lock),!.
clpr_submit_eq([],Exp):-clpr_update_normal_form(Exp, FN), clpr_submit_eq(FN).


inverse(sin(X), M2, X, R):-R is asin(-M2), clpr_non_zero(R), !; R=0, !.
inverse(cos(X), M2, X, R):-R is acos(-M2), clpr_non_zero(R), !; R=0, !.


%%%  DISEQUATIONS 

clpr_submit_le([]):-!.
clpr_submit_le([[]*M]):-!,M<0. 
clpr_submit_le(Exp):- 
    clpr_linear_normal_form(Exp) ->  clpr_submit(1,Exp);
    term_variables(Exp, V),
    clpr_freeze(V,  clpr_submit_le(_,Exp)).

clpr_submit_le(Lock, _):-nonvar(Lock),!.
clpr_submit_le([],Exp):-clpr_update_normal_form(Exp, FN), clpr_submit_le(FN).



/* *** CLP(R) Freeze *** */

%:-attribute('$clpr_frozen'/1).
:-verify_attributes_predicate(clpr_freeze_verif).


clpr_freeze([H|T], G):-!,
    clpr_freeze(T, G),(
    get_atts(H, '$clpr_frozen'(Gs)) ->
        put_atts(H, '$clpr_frozen'((Gs,G)));
        put_atts(H, '$clpr_frozen'(G))).
clpr_freeze([],_).

clpr_freeze_verif(X, Y, G):-
    get_atts(X, '$clpr_frozen'(Gx)),!,(
        float(Y) -> G = Gx;
        get_atts(Y, '$clpr_frozen'(Gy)) -> 
            G  = (Gx,Gy), put_atts(Y,'$clpr_frozen'(true));
        put_atts(Y,'$clpr_frozen'(Gx)), G = true).
clpr_freeze_verif(_, _, true).


/* *** Unification Test *** */

:-verify_attributes_predicate('clpr_index_verif').

clpr_index_verif(X,Y,G):- 
    get_atts(X, '$clpr_index'(Ix)), !, (
    float(Y) -> clpr_unify_with_flt(Ix,Y, G);
    get_atts(Y, '$clpr_index'(Iy)) -> clpr_unify_with_clpr(Ix,Iy,G);
    put_atts(Y, '$clpr_index'(Ix)), G=true ).
clpr_index_verif(_,_,true).


/* *** Portraying *** */

:-portray_attributes_predicate(clpr_get_store).

clpr_get_store(L,C):- set_bip_name(clpr_get_store,2), (
    list(L) ->
	clpr_get_linear_store(L,C1),
	clpr_get_nonlinear_store(L,C2),
	append(C1,C2,C)
    ;
	throw(type_error(list, L))).



clpr_get_nonlinear_store([H|T],L):-!,
    clpr_get_nonlinear_store(T,L2),
   (clpr_get_nonlinear_constraint(H,G2) -> append(L2,G2,L) ; L = L2).
clpr_get_nonlinear_store([],[]).

clpr_get_nonlinear_constraint(X, Gs1):-
    get_atts(X, '$clpr_frozen'(Gs)),
    clpr_nonlinear_pretty_conv(Gs, Gs1).

clpr_nonlinear_pretty_conv((GsL,GsR), Gs1):-!,(
    clpr_nonlinear_pretty_conv(GsL, GsL1) -> 
        (clpr_nonlinear_pretty_conv(GsR, GsR1) ->
           append(GsL1,GsR1,Gs1);
           Gs1=GsL1)
    ;
        clpr_nonlinear_pretty_conv(GsR, Gs1)).

clpr_nonlinear_pretty_conv(clpr_submit_eq(L, Exp), [Exp2]):-
    (var(L) ; get_atts(L, '$clpr_interest'(0))), !,  
     put_atts(L, '$clpr_interest'(1)), 
     Exp2={Exp1=C}, clpr_conv_exp(Exp, Exp1, C).
clpr_nonlinear_pretty_conv(clpr_submit_le(L, Exp), [Exp2]):-
    (var(L) ; get_atts(L, '$clpr_interest'(0))), !,  
     put_atts(L, '$clpr_interest'(1)), 
    Exp2={Exp1=<C}, clpr_conv_exp(Exp, Exp1, C).


clpr_conv_exp([[]*M|T1], Exp, M1):-!,( M1 is -M,
    clpr_conv_exp(T1, Exp, _), !; 
    Exp = 0.0).

clpr_conv_exp([K*M1|T1], Exp, C):- 
    clpr_conv_exp(T1, Exp1, C),!, clpr_conv_exp_1(K, Exp2),
    (M1 < 0.0 -> M is -M1, Exp = Exp1-Exp3; M = M1, Exp = Exp1 + Exp3),
    (M = 1.0 -> Exp3 = Exp2; Exp3 = Exp2 * M).

clpr_conv_exp([K*M1|_], Exp, 0.0):- clpr_conv_exp_1(K, Exp1),
    (M1 < 0 -> M is -M1, Exp= -Exp2; M = M1, Exp = Exp2),
    (M = 1.0 -> Exp2 = Exp1; Exp2 = Exp1 * M).

clpr_conv_exp(NF, Exp):- clpr_conv_exp(NF, Exp1, C),!,
    ( Exp1 == 0.0 -> Exp = C;
    ( clpr_non_zero(C) ->
      ( C < 0.0 , C1 is -C,  Exp = Exp1 + C1; Exp = Exp1 - C);
      Exp = Exp1)).

clpr_conv_exp_1([V1**I1|T1], Exp):-
       ( 
	 generic_var(V1) -> Exp1 = V1;
	 V1=abs(NF) -> clpr_conv_exp(NF, Exp2), Exp1=abs(Exp2);
	 V1=sin(NF) -> clpr_conv_exp(NF, Exp2), Exp1=sin(Exp2);
	 V1=cos(NF) -> clpr_conv_exp(NF, Exp2), Exp1=cos(Exp2);
	 V1=min(NFL, NFR) -> 
	      clpr_conv_exp(NFL, ExpL), clpr_conv_exp(NFR, ExpR), Exp1=min(ExpL, ExpR);
	 V1=max(NFL, NFR) -> 
             clpr_conv_exp(NFL, ExpL), clpr_conv_exp(NFR, ExpR), Exp1=max(ExpL, ExpR);
	 clpr_conv_exp(V1, Exp1)
       ),
    (clpr_conv_exp_1(T1,Exp3) -> (
	I1 = 1.0 -> Exp = Exp1 * Exp3;
	I1 = - 1.0 -> Exp = Exp3 / Exp1;
        Exp = (Exp1**I1 * Exp3)) ; (
    I1 = 1.0 -> Exp = Exp1;
    I1 = -1.0 -> Exp = 1.0 / Exp1;
    Exp = (Exp1**I1))).


/* ******* */



clpr_get_linear_store(V,C):-
    '$call_c_test'('CLPR_Portray'(C1)),
    clpr_extract_constraints_of_interest(V,C1,C2),
    clpr_linear_pretty_conv(C2,C).


clpr_extract_constraints_of_interest(Vars,CI,CO):-
    T = tag(1),
    clpr_extract_constraints_of_interest(Vars,CI,CO,T),
    setarg(1,T,0).


clpr_extract_constraints_of_interest(Vars,CI,CO,Tag):-
    tag_clpr_vars(Vars,Tag),
    clpr_extract_constraints_of_interest2(CI,C1,C2),
    (clpr_extract_untag_var(C1,Vars2), compound(C1) ->
       clpr_extract_constraints_of_interest(Vars2, C2, C3,Tag),
       append(C1,C3,CO);
       CO=C1).

clpr_extract_constraints_of_interest2([H|T], C1, C2):-!,
   (clpr_constraints_of_interest(H) ->
      C1 = [H|T1], C2=T2;
      C1 = T1, C2=[H|T2]
   ),
   clpr_extract_constraints_of_interest2(T,T1,T2).
clpr_extract_constraints_of_interest2([],[],[]).

clpr_constraints_of_interest([(X,_)|_]):-
    get_atts(X,'$clpr_interest'(tag(1))),!.
clpr_constraints_of_interest([_|T]):-
    clpr_constraints_of_interest(T).


tag_clpr_vars([H|T],Tag):-!,
   (float(H);
    put_atts(H,'$clpr_interest'(Tag))),
    tag_clpr_vars(T,Tag).
tag_clpr_vars([],_).


clpr_extract_untag_var([[_|C]|T],Vars):-!,
    clpr_extract_untag_var2(C, V1),
    clpr_extract_untag_var(T,V2),
    append(V2,V1,Vars).
clpr_extract_untag_var([],[]).

clpr_extract_untag_var2([(X,_)|T],Vars):-!,
    (get_atts(X,'$clpr_interest'(tag(1))) ->
      Vars = V1;
      Vars = [X|V1] 
    ),
    clpr_extract_untag_var2(T,V1).
   
clpr_extract_untag_var2([_|T], Vars):-
    clpr_extract_untag_var2(T,Vars).
   

clpr_extract_untag_var2([],[]).


clpr_linear_pretty_conv([[F|H1]|T1],L2):-!,
    clpr_linear_pretty_conv(H1, L, R),
    (L == [] -> L2 = T2;
     H2 =.. [F, L, R], L2 = [{H2}|T2]),
    clpr_linear_pretty_conv(T1,T2).
clpr_linear_pretty_conv([],[]).


clpr_linear_pretty_conv([(X, C)|T], L, R):-!,
    clpr_linear_pretty_conv(T, L1, R),
    (C = 1.0 -> M = X; M = C * X),
    (L1 == [] ->  L = M ; L = L1 + M).
clpr_linear_pretty_conv([R|T], L, R):-!,
   clpr_linear_pretty_conv(T, L, _).
clpr_linear_pretty_conv([], [], 0.0).

       

%/ * *** Interface Predicates *** * /

:-initialization('$call_c'('CLPR_Init')).

%:-attribute('$clpr_index'/1).

clpr_get_index(X, I):-
    get_atts(X, '$clpr_index'(I)), !;
    '$call_c'('CLPR_New_Variable'(I, X)), put_atts(X, '$clpr_index'(I)).

clpr_submit(Type, Exp):-
    '$call_c'('CLPR_New_Constraint'(Type)),
    clpr_add_coef(Exp),!,
    '$call_c_test'('CLPR_Test').

clpr_add_coef([H|T]):-clpr_add_coef(T),(
    H=[]*C -> '$call_c'('CLPR_Add_Const'(C));
    H=[V**1.0]*M , clpr_get_index(V,I),'$call_c'('CLPR_Add_Monom'(I, M))).
clpr_add_coef([]).


clpr_add_cost_coef([H|T], C):-!,clpr_add_cost_coef(T, C2),(
    H=[V**1.0]*M ->
	get_atts(V, '$clpr_index'(I)),
        '$call_c'('CLPR_Add_Monom_to_Cost'(I, M)), 
        C=C2
    ;
        H=[]*C).
clpr_add_cost_coef([],0.0).

clpr_minimize(Exp, Min):-
    clpr_normalize(Exp, 1.0, NF),
    clpr_linear_normal_form(NF),!,
        (   clpr_add_cost_coef(NF,C), 
            '$call_c_test'('CLPR_Minimization'(Min2)),
            Min is  Min2 + C,
	    '$call_c'('CLPR_cost_to_zero'),!
	;
	    '$call_c'('CLPR_cost_to_zero'), fail);
    throw(error('non linear expression'(Exp))).

clpr_unify_with_flt(I, F, G):-
%     '$call_c_test'('CLPR_test_instanciated'(I)), !, G=true;
     '$call_c'('CLPR_New_Constraint'(0)),
     (clpr_non_zero(F) -> '$call_c'('CLPR_Add_Const'(F)); true),
     '$call_c'('CLPR_Add_Monom'(I, -1.0)),
     '$call_c_test'('CLPR_Test'),
     G='$clpr_try_to_detect_instanciation'.

    
clpr_unify_with_clpr(Ix, Iy, '$clpr_try_to_detect_instanciation'):-
     '$call_c'('CLPR_New_Constraint'(0)),
     '$call_c'('CLPR_Add_Monom'(Ix, 1.0)),
     '$call_c'('CLPR_Add_Monom'(Iy, -1.0)),
     '$call_c_test'('CLPR_Test').

'$clpr_try_to_detect_instanciation':-
    '$call_c'('CLPR_Try_to_detect_instanciation').



/* *** */
