/*
    Incremental functions for LP_Solve
    
    Copyright 2002 INRIA, Remy Haemmerle

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "incremental.h"


/*static int sav[400];*/


inclp * 
INC_make_inclp()
{
  inclp * ilp;
  
  CALLOC(ilp, 1);

  ilp->lp = make_lp(0,0);

  ilp->eqs = 0;
  ilp->eqs_alloc = INITIAL_EQ;
  CALLOC(ilp->eq_nr, INITIAL_EQ);
  CALLOC(ilp->eq, INITIAL_EQ);
  CALLOC(ilp->svar,  INITIAL_EQ);
  
  ilp->satisfiable = TRUE;  /* trivialy satisfiable */
  
  return ilp;
}


void 
INC_del_variable(inclp * ilp)
{
  short valid;

#ifndef NO_CHECKING
  if ((ilp->lp->row_end[ilp->lp->rows]!=0)) 
    error("INC_del_variable : try to delete used variable");
#endif
  
  valid = ilp->lp->basis_valid;
  del_constraint(ilp->lp, ilp->lp->rows); 
  if (valid)
    {
      ilp->lp->basis[ilp->lp->bas[ilp->lp->rows + 1]] = FALSE;
      ilp->lp->basis_valid = TRUE;
    }
}

void 
INC_del_inclp(inclp * ilp)
{
  int i;
  for (i=0; i<ilp->eqs ; i++)
    {
      free_mat_rec(ilp->eq[i]);
      free_mat_rec(ilp->svar[i]);
    }
  free(ilp->eq_nr);
  free(ilp->eq);
  free(ilp->svar);
  delete_lp(ilp->lp);
}

void 
INC_add_new_variable(inclp * ilp)
{
  add_empty_constraint(ilp->lp);
}

void 
INC_add_new_constraint(inclp * ilp, short type)
{
#ifndef NO_CHECKING
  if (!(ilp->satisfiable)) 
    error("INC_new_add_constraint : Adding constraint only allowed on a satisfiable problem");
#endif
  
  ilp->satisfiable = FALSE;

  if (type == EQ)
    /* Add a new equation */
    { 
      int eqs;

      INC_inc_eq_size(ilp);
      eqs = ilp->eqs;
      ilp->eq_nr[eqs] = ilp->lp->columns;
      ilp->eq[eqs] = NULL;
      ilp->svar[eqs] = NULL;
      ilp->eqs++;

    }
  else
    /* Add a new inequation */
    {
      add_empty_column(ilp->lp);
    }

}

void 
INC_del_constraint(inclp * ilp)
{
  int eq, i, j;
  REAL v;
  matrec **ptr1, **ptr2;
  matrec *ptr;

  fflush(stdout);

  eq = ilp->eqs-1;

  if ((eq >= 0) && (ilp->eq_nr[eq] == ilp->lp->columns))
  /* the last added constraint is an equation */
    {
    
      if((ilp->svar[eq]!=NULL) && (ilp->eq[eq]!=NULL) && (ilp->eq[eq]->nr.row!=0))
	/* a gauss invert pivoting is necessairy */
	{
	  i = ilp->eq[eq]->nr.row;
	  ptr1=&(ilp->svar[eq]);

	  while (((*ptr1)!=NULL) && ((*ptr1)->nr.row < 0))
	    {
	      j = -((*ptr1)->nr.row + 1);
	      v = (*ptr1)->value;
	      for(ptr = ilp->eq[eq]->next; ptr != NULL; ptr=ptr->next)
		{
		  FIND_MAT(ptr2, ilp->eq[j]->next, ptr->nr.row);
		  if (((*ptr2) != NULL) && ((*ptr2)->nr.row == ptr->nr.row))
		    {
		      (*ptr2)->value +=  ptr->value * v;
		      if ( my_abs((*ptr2)->value) < ilp->lp->epsel)
			FREE_MAT(ptr2);
		    }
		  else
		    INSERT_MAT(*ptr2, ptr->nr.row,  ptr->value * v);
		}
	       FIND_MAT(ptr2, ilp->eq[j], i);
	       EXTRACT_MAT((*ptr1), (*ptr2));
	       (*ptr2)->nr.row = i;
	    }
	  while ((*ptr1)!=NULL)
	    {
	      for(ptr = ilp->eq[eq]->next; ptr != NULL; ptr = ptr->next)
		add_mat(ilp->lp, ptr->nr.row, (*ptr1)->nr.col, 
			(*ptr1)->value * ptr->value);
	      add_mat(ilp->lp, i, (*ptr1)->nr.col, (*ptr1)->value);
	      FREE_MAT(ptr1);
	    }
	}
      free_mat_rec(ilp->eq[eq]);
      ilp->eqs--;
    }
  else
    /* the last added constraint is a  disequation */
    {
      int column = ilp->lp->columns;

      for(i = 1; i <= ilp->lp->rows; i++)
	if(ilp->lp->bas[i] == ilp->lp->rows_alloc + column)
	  {
	    ilp->lp->basis[i] = TRUE;
	    ilp->lp->bas[i]=i;
	  }
	
	for(ptr = ilp->lp->mat[column]; ptr != NULL; ptr = ptr->next)
	  ilp->lp->row_end[ptr->nr.row]--;

	free_mat_rec(ilp->lp->mat[column]);
	ilp->lp->eta_valid = FALSE;
	ilp->lp->sum--;
	ilp->lp->sum2--;
	ilp->lp->columns--;    
    }
  ilp->satisfiable = TRUE;
}


void 
INC_add_to_coef(inclp * ilp, int Var, REAL Value)
{
  int eq, i;
  matrec tmp, *ptr;

#ifndef NO_CHECKING
  if (ilp->satisfiable) 
    error("INC_add_to_coef : Call not allowed on a satisfiable problem");

  if (ilp->lp->scaling_used)
    error("INC_add_to_coef : Scaling not allowed for incremental mode");

  if(Var > ilp->lp->rows || Var < 0)
    error("INC_add_to_coef : Var out of range");
#endif

  eq = ilp->eqs-1;

  if ((eq < 0) || (ilp->eq_nr[eq] != ilp->lp->columns)) eq++;
 
  if(ilp->lp->row_end[Var] == 0)
    {
      /* look for if Var is a ref variable */
      for(i=0; i<eq; i++)
	if (ilp->eq[i]->nr.row == Var) break;
    }
  else i = eq;

  if (i < eq) 
    /* Var is a ref variable */
    ptr =  ilp->eq[i]->next;
  else
    {
      /* Var is not a ref variable */
      tmp.nr.row=Var;
      tmp.value=-1;
      tmp.next = NULL;
      ptr = &tmp;
    }
     

  if ((eq < ilp->eqs))
   {
      /* The current new constraint is an equation */
      for(; ptr != NULL; ptr = ptr->next)
	{
	  matrec ** ptr1;
	  
	  FIND_MAT(ptr1, ilp->eq[eq], ptr->nr.row);
	  
	  if( ((*ptr1) != NULL) && ((*ptr1)->nr.row ==  ptr->nr.row)) 
	    /* there is an existing entry */
	    {
	      REAL result;
	      result = ((*ptr1)->value -= Value * ptr->value);
	      if (my_abs(result) <= ilp->lp->epsilon)
		FREE_MAT(ptr1);
	    }
	  else
	    /* no existing entry */
	    {
	      INSERT_MAT(*ptr1, (ptr->nr.row), (-Value * ptr->value));
	    }
	}
    }
  else
    {
      /* The current new constraint is a disequation */
      for(; ptr != NULL; ptr = ptr->next)
	add_mat(ilp->lp, ptr->nr.row, ilp->lp->columns,-Value * ptr->value );
    }

}

short 
INC_minimize(inclp * ilp, REAL *  Value)
{
#ifndef NO_CHECKING
  if (! ilp->satisfiable) 
    error("INC_minimize : Call not allowed on a non satisfiable problem");

  if (ilp->lp->scaling_used)
    error("INC_minimize : Scaling not allowed for incremental mode");
#endif

  switch(solve(ilp->lp))
    {
    case OPTIMAL :
      *Value = (REAL) ilp->lp->best_solution[0];
      return TRUE;
    case INFEASIBLE :
      break;
    default :
      error("INC_test_satisfiabile : bad call");
    }
  return FALSE;
}

short 
INC_test_satisfiable(inclp * ilp, int * svar)
{
  int eq;
  matrec ** ptr ;

  eq = ilp->eqs - 1; 

  /*  sav[eq]=ilp->lp->rows;*/

  if (ilp->satisfiable) return TRUE;

  if ((eq >= 0) && (ilp->eq_nr[eq] >= ilp->lp->columns))
    /* the current new constraint is an equation */
    {
      ptr = &ilp->eq[ilp->eqs - 1];
      if (((*ptr) == NULL) || (((*ptr)->nr.row == 0) && ((*ptr)->next == NULL)))
	{
	  if ( ((*ptr) == NULL) || ((*ptr)->value == 0))
	    return TRIVIAL;
	  else
	    return FALSE;
	}

      if (!INC_gauss_pivot(ilp, -1)) return FALSE;
    }
  else
    {
      ptr = &(ilp->lp->mat[ilp->lp->columns]) ;
      if ( ((*ptr) == NULL) || (((*ptr)->nr.row == 0) && ((*ptr)->next == NULL)))
	{  
	  if ( ((*ptr) == NULL) || ((*ptr)->value >= 0))
	    return TRIVIAL;
	  else
	    return FALSE;
	}
    }

  switch(solve(ilp->lp))
    {
    case OPTIMAL :
      break;
    case UNBOUNDED :
      return FALSE;
    default :
      error("INC_test_satisfiabile : bad call");
    }

  ilp->satisfiable = TRUE;
  return TRUE;
}


void 
INC_cost_to_zero(inclp * ilp)
{
  int i;

  for(i=0; i <= ilp->lp->rows; i++)
    set_rh(ilp->lp, i, 0.0);
}


void 
INC_add_to_cost_coef(inclp * ilp, int Var, REAL Value)
{
  int eq, i;
  matrec * ptr;

#ifndef NO_CHECKING
  if (!ilp->satisfiable) 
    error("INC_add_to_cost_coef : Call not allowed on a non satisfiable problem");

  if (ilp->lp->scaling_used)
    error("INC_add_to_cost_coef : Scaling not allowed for incremental mode");

  if(Var > ilp->lp->rows || Var < 0)
    error("INC_add_to_cost_coef : Var out of range");
#endif
  
  eq = ilp->eqs;

  if(ilp->lp->row_end[Var] == 0)
    {
      /* look for if Var is a ref variable */
      for(i=0; i<eq; i++)
	if (ilp->eq[i]->nr.row == Var) break;
      if (i < eq) 
	{
	  /* Var is a ref variable */
	  ptr =  ilp->eq[i]->next;
	  
	  for(; ptr !=NULL; ptr = ptr->next)
	    add_rh(ilp->lp, ptr->nr.row, -Value * ptr->value );
	  return;
	}
    }

  add_rh(ilp->lp, Var, Value); 
}



void 
INC_inc_eq_size(inclp * ilp)
{
  if(ilp->eqs >= ilp->eqs_alloc)
    {
       ilp->eqs_alloc = ilp->eqs_alloc*1.5;
       REALLOC(ilp->eq_nr,      ilp->eqs_alloc);
       REALLOC(ilp->eq,         ilp->eqs_alloc);
       REALLOC(ilp->svar,       ilp->eqs_alloc);
    }
}


short 
INC_gauss_pivot(inclp * ilp, int pivot)
{
  matrec ** ptr1, ** ptr2, ** ptr3, *ptr;
  int eq, i;

  eq = ilp->eqs - 1;

#ifndef NO_CHECKING
  if (ilp->satisfiable) 
    error("INC_gauss_pivot : Call not allowed on a satisfiable problem");

  if (ilp->lp->scaling_used)
    error("INC_gauss_pivot : Scaling not allowed for incremental mode");

  if ((eq < 0) && (ilp->eq_nr[eq] < ilp->lp->columns))
    error("INC_gauss_pivot : the current new constraint is not an equation");
#endif

  if (pivot == -1)
    /* no imposed pivot variable ; it must be choosen */
    {
      for(ptr1 = &(ilp->eq[eq]); (*ptr1) != NULL; ptr1 = &((*ptr1)->next))
	if (((*ptr1)->nr.row != 0) && (my_abs((*ptr1)->value)>ilp->lp->epsel))
	  {
	    pivot = (*ptr1)->nr.row;
	    break;
	  }
      if (pivot == -1)
	/* equation without variable */
	{
	  if (ilp->eq[eq] == NULL)
	    /* empty constraint ; tautologie*/
	    return TRUE; /* trivialy satisfiable */
	  if (ilp->eq[eq]->nr.row == 0) 
	    /* the equation has a non nul rhs*/
	    return FALSE; /* trivialy not satisfiable */
	}  
    }
  else
    /* verify that the given pivot is not nul */
    {
#ifndef NO_CHECKING
      if (pivot > ilp->lp->rows || pivot <= 0)
	error("INC_gauss_pivot : pivot out of range");
#endif
      
      FIND_MAT(ptr1, ilp->eq[eq], pivot);
      if (((*ptr1) == NULL) || ((*ptr1)->nr.row != pivot))
	return FALSE;
    }
  
  ptr = NULL;
  /* place the pivot in the beginning of the list ilp->eq*/ 
  EXTRACT_MAT((*ptr1), ptr);
  EXTRACT_MAT(ptr, ilp->eq[eq]);

  INC_scale_eq(ilp); /* scale equation before pivotting */
  
  ilp->svar[eq] = extract_row(ilp->lp, pivot);
  


  for(i = 0; i < eq; i++)
    {
      FIND_MAT(ptr1, ilp->eq[i]->next, pivot);
      if (((*ptr1) != NULL) && ((*ptr1)->nr.row == pivot)){
	EXTRACT_MAT((*ptr1), ilp->svar[eq]);
	ilp->svar[eq]->nr.col=-(i+1);
      }
    }

  for(ptr1 = &(ilp->svar[eq]); (*ptr1) != NULL; ptr1 = & ((*ptr1)->next))
     for(ptr2=&(ilp->eq[eq]->next); (*ptr2) != NULL; ptr2=&((*ptr2)->next))
       if ((*ptr1)->nr.row >= 0)
	 {
	   add_mat(ilp->lp, (*ptr2)->nr.row, (*ptr1)->nr.col,
		   - (*ptr2)->value * (*ptr1)->value);
	 }
       else
	 {
	   FIND_MAT(ptr3, (ilp->eq[-((*ptr1)->nr.row+1)])->next, (*ptr2)->nr.row);
	   if ((*ptr3 != NULL) && ((*ptr3)->nr.row == (*ptr2)->nr.row))
	     {
	       (*ptr3)->value -= (*ptr2)->value * (*ptr1)->value;
	       if ( my_abs((*ptr3)->value) < ilp->lp->epsel)
		 FREE_MAT(ptr3);
	     }
	   else
	     INSERT_MAT(*ptr3, (*ptr2)->nr.row, 
			- (*ptr2)->value * (*ptr1)->value);
	 }

  if (ilp->lp->bas[pivot] != pivot)
    {
      ilp->lp->basis[ilp->lp->bas[pivot]]= FALSE;
      ilp->lp->basis[pivot] = TRUE;
      ilp->lp->bas[pivot] = pivot;
    }

  return TRUE;
}


void 
INC_scale_eq(inclp * ilp)
{
  int eq;
  matrec * ptr;
  REAL tmp;

  eq = ilp->eqs - 1;
  ptr = ilp->eq[eq];

#ifndef NO_CHECKING
  if (ilp->satisfiable) 
    error("INC_scale_eq : Call not allowed on a satisfiable problem");

  if ((eq < 0) && (ilp->eq_nr[eq] < ilp->lp->columns))
    error("INC_scale_eq : the current new constraint is not an equation");

  if ((ptr == NULL) || ((ptr->nr.row == 0) && (ptr->next == NULL)))
    error("INC_scale_eq : pivot not ready");
#endif

  tmp = ptr->value;
 
  ptr = ptr->next;
  for(; ptr != NULL; ptr = ptr->next)
      ptr->value /= tmp; 
}


void 
INC_print_inclp(inclp * ilp)
{
  int i, j, dis, eq, w, h;
  REAL * fatmat;
  matrec * ptr;
  REAL mult;
  
  w = ilp->lp->rows + 2;
  h = (ilp->eqs + ilp->lp->columns+1);

  dis = 1; eq=0;
  CALLOC(fatmat, w * h *2);

  for(j=0; j <= ilp->lp->rows; j++)
    fatmat[j] = ilp->lp->orig_rh[j];
  
  i = w;
  while(( eq < ilp->eqs) || (dis <= ilp->lp->columns))
    {
      if (( eq < ilp->eqs) && (ilp->eq_nr[eq] < dis))
	/* The constraint is an equation */
	{
	  fatmat[i + w - 1] = EQ;
	  ptr = ilp->eq[eq];

	  if ( ptr != NULL){
	    mult = ptr->value;
	    fatmat[i + ptr->nr.row] = 1.0;
	    ptr = ptr->next;
	  }
	  eq++;
	}
      else
	/* The constraint is a disequation */
	{
	  mult = 1.0;
	  fatmat[i + w - 1] = LE;
	  ptr = ilp->lp->mat[dis];
	  dis++;
	}

      for(; ptr != NULL; ptr = ptr->next)
	fatmat[i + ptr->nr.row] = ptr->value;

      i += w;
    } 
  
  printf((fatmat[w-1] == GE) ? "Max " : "Min ");
  for(i = 1; i < w - 1; i++)
    printf("%7.3f ",  fatmat[i]);
   printf("   %7.3f\n", fatmat[0]);
   
  for(j = 1; j < h; j++ )
    {
       printf("    ");
      for(i = 1; i < w - 1; i++)
	printf("%7.3f ",  fatmat[j*w+i]);
      printf("%s %7.3f\n", ((fatmat[j*w+w-1] == EQ) ? "==" : "=<")
	     , fatmat[j*w]);
    }
   
  free(fatmat);
}

void 
add_empty_constraint(lprec * lp)
{
  
  lp->rows++;
  lp->sum++;
  inc_row_space(lp);

  lp->col_no_alloc[lp->rows]=0;
  lp->row_end[lp->rows]=0;
  lp->col_no[lp->rows] = NULL;
  
  lp->ch_sign[lp->rows] = FALSE;
  lp->orig_upbo[lp->rows] = 0;
  lp->orig_lowbo[lp->rows] = 0;
  lp->orig_rh[lp->rows] = 0;

  lp->bas[lp->rows] = lp->rows;
  lp->basis[lp->rows] = TRUE;
  lp->lower[lp->rows] = TRUE;   
  lp->eta_valid = FALSE;
}


void 
add_empty_column(lprec *lp)
{
 
  lp->columns++;
  lp->sum++;
  lp->sum2++;
  inc_col_space(lp);
  lp->mat[lp->columns] =  NULL;
  lp->orig_lowbo[lp->sum2] = 0;
  lp->orig_upbo[lp->sum2] = lp->infinite;
  lp->lower[lp->sum2] = TRUE;
  lp->basis[lp->sum2] = FALSE;
  lp->must_be_int[lp->sum2] = FALSE;
  if(lp->names_used)
    sprintf(lp->col_name[lp->columns], "var_%d", lp->columns);
 
  
}

REAL 
add_mat(lprec * lp, int Row, int Column, REAL Value)
{

  Value +=  mat_elm(lp, Row, Column);
  set_mat(lp, Row, Column, Value);
  
  return Value;
}

REAL 
get_rh(lprec *lp, int Row)
{
  REAL value;
  
  value = lp->orig_rh[Row];

  if(value == 0.0) return 0.0;

  if (lp->scaling_used)
    value /= lp->scale[Row];

  if((lp->ch_sign[Row]) || ((Row == 0) && (!lp->maximise)))
    return -value;
  else
    return value;
}

REAL 
add_rh(lprec * lp, int Row, REAL Value)
{
  if (Row == 0) Value = -Value;

  Value += get_rh(lp, Row);
  set_rh(lp, Row, Value);
  return Value;
}


matrec * 
extract_row(lprec * lp, int Row)
/* 
   Warning : !! In the extracted Row the field returned by this function 
   row_nr of each element are in fact the columns numeros of the element 
*/   
{
  matrec ** ptr1;
  matrec * ptr;
  int i;

#ifndef NO_CHECKING
  if(Row > lp->rows || Row < 0)
    error("extract_row : Row out of range");
#endif
  
  ptr = NULL;
  
  for(i = 0; i < lp->row_end[Row]; i++)
    {
      FIND_MAT(ptr1, lp->mat[lp->col_no[Row][i]], Row);
      EXTRACT_MAT((*ptr1), ptr); 
      ptr->nr.col = lp->col_no[Row][i]; 
    }
  
  lp->row_end[Row] = 0;
	
  return ptr;
} 

/* *** Debug Tests *** */
/*
int _cmptr = 0;

void 
verify_row_end_coherence(lprec * lp, char * text)
{
  int i,j;
  matrec ** ptr;


  for(i=0; i<= lp->rows; i++)
    for(j=0; j<lp->row_end[i]; j++)
      {
	FIND_MAT(ptr, lp->mat[lp->col_no[i][j]], i);
	if (((*ptr) == NULL) || ((*ptr)->nr.row != i))
	  {
	    printf("%s\tlp->row_end_valid = %i\tlp->rows = %i\ti = %i\tcmptr = %i\n", (text == NULL) ? "" : text, lp->row_end_valid, lp->rows, i, _cmptr);
	  }
      }
  
}

void 
coherency(inclp * ilp)
{
  int i, j;
  matrec * ptr;
  matrec ** ptr1;

  for(i=1; i<=ilp->lp->rows; i++)
    {
      j = ilp->lp->bas[i];

      if (0<=j && j <=ilp->lp->rows && j==i && ilp->lp->basis[i] == TRUE)
	continue;
      if (ilp->lp->rows_alloc<j && j<=ilp->lp->rows_alloc+ilp->lp->columns &&
	  ilp->lp->basis[i] == FALSE && ilp->lp->basis[j] == TRUE)
	continue;
	{
	  printf("Pbr ilp->lp->bas[%i]\n", i);
	}
    }

  j = ilp->lp->row_end[0];

  for(i=1; i<= ilp->lp->columns; i++)
    {
      FIND_MAT(ptr1, ilp->lp->mat[i], 0);
      if(((*ptr1) != NULL) && ((*ptr1)->nr.row == 0)) j--;
    }
  if (j != 0 ) 
    {
      printf("PBR toto -- %i -- %i --\n",  ilp->lp->row_end[0], ilp->lp->row_end[0]-j );
    }
 
  for(i=0; i <= ilp->lp->rows; i++)
    if ((ilp->lp->row_end[i] < 0) || (ilp->lp->row_end[i] > ilp->lp->columns))
	printf("Prb ilp->lp->row_end[%i] = %i\n", i, ilp->lp->row_end[i]);


  for(i=0; i<=ilp->lp->columns; i++)
    for(ptr=ilp->lp->mat[i]; ptr !=NULL; ptr=ptr->next)
      if (ptr->nr.row > ilp->lp->rows)
	{
	  printf("\tlp_coherency\tDISEQ\tlp->rows = %i\tr=%i\tc=%i\t(%i)\n", ilp->lp->rows, ptr->nr.row, i, _cmptr);
	}
  for(i=0; i<ilp->eqs; i++)
     for(ptr=ilp->eq[i]; ptr !=NULL; ptr=ptr->next)
      if (ptr->nr.row > ilp->lp->rows)
	{
	  printf("\tlp_coherency\tEQUA\tlp->rows = %i\tr=%i\tc=%i\t(%i)\n", ilp->lp->rows, ptr->nr.row, i, _cmptr);
	}

}

*/
